#ifndef FHGFSOPS_H_
#define FHGFSOPS_H_

#include <include_closed/app/App.h>
#include <include_open/common/threading/Open_Mutex.h>
#include <include_open/common/Open_Common.h>
#include <include_closed/net/filesystem/FhgfsOpsRemoting.h>
#include <include_closed/filesystem/FsDirInfo.h>
#include <include_closed/filesystem/FsFileInfo.h>
#include <include_closed/filesystem/FhgfsOpsHelper.h>
#include <include_closed/toolkit/NoAllocBufferStore.h>
#include <open/common/Common.h>
#include "FhgfsOps_versions.h"


#include <linux/fs.h>
#include <linux/dcache.h>
#include <linux/namei.h>
#include <linux/pagemap.h>
#include <linux/vfs.h>
#include <linux/mount.h>
#include <linux/mpage.h>
#include <linux/writeback.h>


#ifndef SEEK_SET
#define SEEK_SET  0  /* seek relative to beginning of file */
#define SEEK_CUR  1  /* seek relative to current file position */
#define SEEK_END  2  /* seek relative to end of file */
#endif


#define FHGFS_STAT_BLOCKBITS    PAGE_SHIFT /*(16)*/ /* debug value */
#define FHGFS_STAT_BLOCKSIZE    (1 << FHGFS_STAT_BLOCKBITS)


// forward declaration
struct App;



// functions

int FhgfsOps_registerFilesystem(void);
int FhgfsOps_unregisterFilesystem(void);

extern struct dentry* FhgfsOps_lookup(struct inode* parentDir, struct dentry* dentry,
   struct nameidata* nameidata);
extern int FhgfsOps_revalidate(struct dentry* dentry, struct nameidata* nameidata);
extern int FhgfsOps_deleteDentry(struct dentry* dentry);

extern loff_t FhgfsOps_llseek(struct file *file, loff_t offset, int origin);

extern int FhgfsOps_getattr(struct vfsmount* mnt, struct dentry* dentry, struct kstat* kstat);
extern int FhgfsOps_setattr(struct dentry* dentry, struct iattr* iattr);
extern int FhgfsOps_permission(struct inode* inode, int mask, struct nameidata* nameidata);

extern int FhgfsOps_mknod(struct inode* dir, struct dentry* dentry, int mode, dev_t dev);

extern int FhgfsOps_mkdir(struct inode* dir, struct dentry* dentry, int mode);
extern int FhgfsOps_rmdir(struct inode* dir, struct dentry* dentry);
extern int FhgfsOps_opendirIncremental(struct inode* inode, struct file* file);
extern int FhgfsOps_readdirIncremental(struct file* file, void* buf, filldir_t filldir);
extern int FhgfsOps_releasedir(struct inode* inode, struct file* file);

extern int FhgfsOps_create(struct inode* dir, struct dentry* dentry, int mode,
   struct nameidata* nameidata);
extern int FhgfsOps_unlink(struct inode* dir, struct dentry* dentry);
extern int FhgfsOps_open(struct inode* inode, struct file* file);
extern int FhgfsOps_release(struct inode* inode, struct file* file);
extern int FhgfsOps_fsync(struct file* file, struct dentry* dentry, int datasync);

extern int FhgfsOps_symlink(struct inode* dir, struct dentry* dentry, const char* to);
extern int FhgfsOps_readlink(struct dentry* dentry, char __user* buf, int size);

extern int FhgfsOps_rename(struct inode* inodeDirFrom, struct dentry* dentryFrom,
   struct inode* inodeDirTo, struct dentry* dentryTo);

extern int FhgfsOps_link(struct dentry* dentryFrom, struct inode* inode, struct dentry* dentryTo);

extern void FhgfsOps_truncate(struct inode* inode); // i_size of the inode is set to the desired size

extern ssize_t FhgfsOps_read(struct file* file, char __user *buf, size_t size,
   loff_t* offsetPointer);
extern ssize_t FhgfsOps_write(struct file* file, const char __user *buf, size_t size,
   loff_t* offsetPointer);
extern int FhgfsOps_mmap(struct file *, struct vm_area_struct *);

extern int FhgfsOps_readpage(struct file *file, struct page *page);
extern int FhgfsOps_writepage(struct page *page, struct writeback_control *wbc);
extern int FhgfsOps_prepare_write(struct file *file, struct page *page, unsigned from,
   unsigned to);
extern int FhgfsOps_commit_write(struct file *file, struct page *page, unsigned from,
   unsigned to);
extern sector_t FhgfsOps_bmap(struct address_space *, sector_t);
extern ssize_t FhgfsOps_directIO(int rw, struct kiocb *iocb, const struct iovec *iov, loff_t pos,
   unsigned long nr_segs);

extern void FhgfsOps_putSuper(struct super_block* sb);
extern void FhgfsOps_killSB(struct super_block* sb);


extern struct inode* __FhgfsOps_newInode(struct super_block* sb, struct kstat* kstat, dev_t dev);
extern int __FhgfsOps_instantiateInode(struct dentry* dentry, char* path);
extern int __FhgfsOps_fillSuper(struct super_block* sb, void* rawMountOptions, int silent);
extern int __FhgfsOps_readpage_sync(struct file* file, struct page *page);
extern int __FhgfsOps_writepage_sync(struct file* file, struct page* page, unsigned pageOffset,
   unsigned count);
extern int __FhgfsOps_writepage_async(struct page* page, unsigned from, unsigned to);
extern int __FhgfsOps_refreshInodeUnlocked(struct dentry* dentry, char* path);
extern int __FhgfsOps_refreshInode(struct dentry* dentry, char* path);

extern int __FhgfsOps_initApp(struct super_block* sb, char* rawMountOptions);
extern void __FhgfsOps_uninitApp(App* app);


// inliners
//static inline char* __FhgfsOps_fullpathResolveToPage(
//   struct dentry* dentry, struct vfsmount* mnt, char** outPageBuf);
static inline char* __FhgfsOps_pathResolveToStoreBuf(NoAllocBufferStore* bufStore,
   const struct dentry* dentry, char** outStoreBuf);
static inline void __FhgfsOps_applyStatDataToInode(struct kstat* kstat, struct inode* outInode);
static inline void __FhgfsOps_lockFhgfs(App* app);
static inline void __FhgfsOps_unlockFhgfs(App* app);

// getters & setters
static inline App* FhgfsOps_getApp(struct super_block* sb);
static inline void FhgfsOps_setApp(struct super_block* sb, App* app);
static inline FsObjectInfo* FhgfsOps_getObjectInfo(struct file* file);
static inline void FhgfsOps_setObjectInfo(FsObjectInfo* objectInfo, struct file* file);
static inline FsDirInfo* FhgfsOps_getDirInfo(struct file* file);
static inline void FhgfsOps_setDirInfo(FsDirInfo* dirInfo, struct file* outFile);
static inline FsFileInfo* FhgfsOps_getFileInfo(struct file* file);
static inline void FhgfsOps_setFileInfo(FsFileInfo* fileInfo, struct file* outFile);



// Note: Not needed, and declaration was changed in kernel 2.6.25
///**
// * Allocates a page that must be free_page'd by the caller.
// * Note: unlike the other path resolve method, this version retrieves the full path (including
// * the path to the mount point)
// *
// * @param outPageBuf the page that must be free_page'd
// * @return points to some location inside the pageBuf
// */
//char* __FhgfsOps_fullpathResolveToPage(struct dentry* dentry, struct vfsmount* mnt,
//   char** outPageBuf)
//{
//   char* resolvedPath;
//
//   *outPageBuf = (char*)__get_free_page(GFP_NOFS);
//
//   resolvedPath = d_path(dentry, mnt, *outPageBuf, PAGE_SIZE);
//
//   return resolvedPath;
//}


/*
 * Constructs the path from the root dentry (of the mount-point) to an arbitrary hashed dentry.
 *
 * Note: Acquires a buf from the pathBufStore that must be released by the caller.
 * Note: This is safe for paths larger than bufSize, but returns -ENAMETOOLONG in that case
 *
 * @outBuf the buf that must be returned to the pathBufStore of the app
 * @return some offset within the outStoreBuf
 */
char* __FhgfsOps_pathResolveToStoreBuf(NoAllocBufferStore* bufStore, const struct dentry* dentry,
   char** outStoreBuf)
{
   const ssize_t storeBufLen = NoAllocBufferStore_getBufSize(bufStore);
   ssize_t bufLenLeft; // remaining length
   char* currentBufStart;
   int nameLen;

   bufLenLeft = storeBufLen;
   *outStoreBuf = NoAllocBufferStore_waitForBuf(bufStore);

   currentBufStart = (*outStoreBuf) + bufLenLeft;

   *--currentBufStart = '\0';
   bufLenLeft--;

   spin_lock(&dcache_lock);

   while(!IS_ROOT(dentry) )
   {
      nameLen = dentry->d_name.len;
      bufLenLeft -= nameLen + 1;

      if (bufLenLeft < 0)
         goto err_too_long_unlock;

      currentBufStart -= nameLen;

      memcpy(currentBufStart, dentry->d_name.name, nameLen);

      *--currentBufStart = '/';

      dentry = dentry->d_parent;
   }

   spin_unlock(&dcache_lock);

   // return "/" instead of an empty string for the root directory
   if(bufLenLeft == (storeBufLen-1) ) // -1 for the terminating zero
   { // dentry is (mount) root directory
      *--currentBufStart = '/';
   }

   return currentBufStart;

err_too_long_unlock:
   spin_unlock(&dcache_lock);

   return ERR_PTR(-ENAMETOOLONG);
}

void __FhgfsOps_applyStatDataToInode(struct kstat* kstat, struct inode* outInode)
{
   // remote attribs (received from nodes)
   outInode->i_mode = kstat->mode;
   outInode->i_uid = kstat->uid;
   outInode->i_gid = kstat->gid;
   outInode->i_atime = kstat->atime;
   outInode->i_mtime = kstat->mtime;
   outInode->i_ctime = kstat->ctime;
   outInode->i_size = kstat->size;
   outInode->i_nlink = kstat->nlink;

   // local attribs
   outInode->i_blkbits = FHGFS_STAT_BLOCKBITS;
   outInode->i_blocks = 0;
   if(outInode->i_size)
   {
      // note: i_blocks measures 512-byte units of the file size (see man stat(2) )

      // note: do_div(n64, base32) assigns the result to n64 and returns the remainder!
      outInode->i_blocks = outInode->i_size-1;
      do_div(outInode->i_blocks, 512);
      outInode->i_blocks++; // things that have a positive size always require at least one block
   }


//#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,19)
#ifdef KERNEL_HAS_IBLKSIZE
   outInode->i_blksize = FHGFS_STAT_BLOCKSIZE;
#endif // KERNEL_HAS_IBLKSIZE
//#endif // LINUX_VERSION_CODE

}


void __FhgfsOps_lockFhgfs(App* app)
{
   Mutex* fhgfsOpsMutex = App_getFhgfsOpsMutex(app);

#ifdef LOG_DEBUG_MESSAGES
//   Logger* log = App_getLogger(Program_getApp() );
//   const char* logContext = "__FhgfsOps_lockFhgfs";
//   Logger_logFormatted(log, 5, logContext, "Waiting for BigKernelLock(BKL)");
#endif

   //lock_kernel(); // B-K-LOCK
   Mutex_lock(fhgfsOpsMutex);

#ifdef LOG_DEBUG_MESSAGES
//   Logger_logFormatted(log, 5, logContext, "Got BigKernelLock(BKL)");
#endif
}

void __FhgfsOps_unlockFhgfs(App* app)
{
   Mutex* fhgfsOpsMutex = App_getFhgfsOpsMutex(app);

#ifdef LOG_DEBUG_MESSAGES
//   Logger* log = App_getLogger(Program_getApp() );
//   const char* logContext = "__FhgfsOps_unlockFhgfs";
//   Logger_logFormatted(log, 5, logContext, "Releasing BigKernelLock(BKL)");
#endif

   Mutex_unlock(fhgfsOpsMutex);
   //unlock_kernel(); // B-K-UNLOCK

#ifdef LOG_DEBUG_MESSAGES
//   Logger_logFormatted(log, 5, logContext, "Released BigKernelLock(BKL)");
#endif
}


App* FhgfsOps_getApp(struct super_block* sb)
{
   return (App*)sb->s_fs_info;
}

void FhgfsOps_setApp(struct super_block* sb, App* app)
{
   sb->s_fs_info = app;
}


FsObjectInfo* FhgfsOps_getObjectInfo(struct file* file)
{
   return (FsObjectInfo*)file->private_data;
}

void FhgfsOps_setObjectInfo(FsObjectInfo* objectInfo, struct file* file)
{
   file->private_data = objectInfo;
}

FsDirInfo* FhgfsOps_getDirInfo(struct file* file)
{
   return (FsDirInfo*)file->private_data;
}

void FhgfsOps_setDirInfo(FsDirInfo* dirInfo, struct file* outFile)
{
   outFile->private_data = dirInfo;
}

FsFileInfo* FhgfsOps_getFileInfo(struct file* file)
{
   return (FsFileInfo*)file->private_data;
}

void FhgfsOps_setFileInfo(FsFileInfo* fileInfo, struct file* outFile)
{
   outFile->private_data = fileInfo;
}


#endif /*FHGFSOPS_H_*/
